//////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009 - 2009 RECITAL CORPORATION Inc.
// As an unpublished licensed proprietary work.
// All rights reserved worldwide.
//////////////////////////////////////////////////////////////////////////////
//
//      MODULE          : fifiagent.prg
//
//      PURPOSE         : Replication Data Agent
//
//      AUTHOR          : Recital Corporation
//
//      DATE            : 23-oct-2009
// 
//////////////////////////////////////////////////////////////////////////////

// defines
#define FIFO_ALTER 			5099
#define FIFO_CONFLICT 		1409
#define FIFO_CREATE 		5017
#define FIFO_DELETE 		5022
#define FIFO_INSERT   		5042
#define FIFO_PACK   		33
#define FIFO_QUEUEMARKER	1413
#define FIFO_RECALL   		36
#define FIFO_UPDATE 		5082
#define FIFO_ZAP    		82

// Setup Environment
set debug on
set exclusive off

// define public variables
public  p_publicationid = val(getenv("DB_PUBLICATIONID"))
public p_dataobject

// Open the replication queue
fifo_opendata()

// Open the queue
m_fifo_fd=fopen("fifo://"+getenv("DB_DATADIR")+"replication/fifo.pipe")
if (m_fifo_fd < 0) 
   debug(strerror(error()))
   return
endif

// Proccess the message queue
do while .t.
   // Check for new messages
   m_message = fgets(m_fifo_fd)
   do while (!empty(m_message))
	   // decode the XML message into a dynamic array
	   m_objmessage = xml_decode(m_message)  
	   // process the message
       fifo_process_message(m_objmessage, m_message)
       m_message = fgets(m_fifo_fd)
   enddo
   sleep 1
enddo

// Close open files
fclose(m_fifo_fd)
close database

//////////////////////////////////////////////////////////////////////////////
//
// Functions
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// fifo_existing()
//
// Parameters 
//
//  The message strored in a dynamic array
//
// Summary
//
// Returns .t. if existing transaction is in the queue
// that can be merged into the new one.
//
function fifo_existing(m_objmessage)
	m_rc = .f.
	if  val(m_objmessage._rep_ct) = FIFO_UPDATE
        m_key =  fifo_getkey(m_objmessage)
        seek m_key
        scan while &(key(order())) = m_key 
            if queue.cmdtype = FIFO_ZAP  or queue.cmdtype = FIFO_PACK
	            continue
	        elseif queue.cmdtype = FIFO_INSERT  or queue.cmdtype = FIFO_UPDATE
	            m_rc = .t.
		        exit
	        elseif queue.cmdtype = FIFO_CONFLICT  or queue.cmdtype = FIFO_QUEUEMARKER
		        exit
	        endif
        endscan
    endif
return m_rc

//////////////////////////////////////////////////////////////////////////////
//
// fifo_getkey()
//
// Parameters 
//
//  The message strored in a dynamic array
//
// Summary
//
// Return a character string representing a searchable key
// in the queue table on the cmdstate index tag
//
function fifo_getkey(m_objmessage)
    m_klen = len(queue.trankey)
    m_key = "W"+left(rpad(m_objmessage._rep_tn,m_klen),m_klen-strlen(m_objmessage._rep_sn))+m_objmessage._rep_sn
return m_key

//////////////////////////////////////////////////////////////////////////////
//
// fifo_merge()
//
// Parameters 
//
//  The message strored in a dynamic array
//
// Summary
//
// Merge existing transaction in the replication queue 
// into the transaction.
//
function fifo_merge(m_objmessage, m_xmlexisting)
   foreach m_xmlexisting as _name => _data
      if type("m_objmessage.&_name") = "U"
	    m_objmessage.&_name = _data
      elseif lower(_name) = "_rep_tt"
	    m_objmessage.&_name = _data
      elseif lower(_name) = "_rep_ct"
	    m_objmessage.&_name = _data
      endif
   endfor
return

//////////////////////////////////////////////////////////////////////////////
//
// fifo_builddata()
//
// Parameters 
//
//  The message strored in a dynamic array
//  The message strored in a character string
//
// Summary
//
// Update the data dynamic array with the transaction data 
//
function fifo_builddata(m_objmessage,m_message)
   foreach p_dataobject as _name => _data
	   do case
	   case _name = "PUBLISHER"
           p_dataobject.&_name = m_objmessage._rep_pn 
	   case _name = "PUBLICATION"
           p_dataobject.&_name = strzero(p_publicationid,2)
	   case _name = "TRANID"
           p_dataobject.&_name = val("0x"+substr(cursyncnum(),3))+1
	   case _name = "DATA"
           p_dataobject.&_name = m_message
	   case _name = "CMDTYPE"
           p_dataobject.&_name = val(m_objmessage._rep_ct)
	   case _name = "INSERTDATE"
           p_dataobject.&_name = datetime()
	   case _name = "CMDSTATE"
           p_dataobject.&_name = "W"
	   case _name = "TRANKEY"
           p_dataobject.&_name = substr(fifo_getkey(m_objmessage),2)
	   case _name = "TRANTYPE"
           p_dataobject.&_name = "Q"
	   endcase
   endfor
return

//////////////////////////////////////////////////////////////////////////////
//
// fifo_process_message()
//
// Parameters 
//
//  The message strored in a dynamic array
//  The message strored in a character string
//
// Summary
//
// The main message processing  
//
function fifo_process_message(m_objmessage,m_message)

    // Check for existing non processed transaction
	if (fifo_existing(m_objmessage))
	    fifo_merge(m_objmessage, xml_decode(queue.data))
	    m_rowid = rowid
		m_xmlmerged = xml_encode(m_objmessage)
        update replication!queue set data = m_xmlmerged where rowid = m_rowid
	else
	    fifo_builddata(m_objmessage,m_message)
        insert into replication!queue from name p_dataobject
	endif
	// flush network cache
	skip 0
return

//////////////////////////////////////////////////////////////////////////////
//
// fifo_opendata()
//
// Summary
//
// Open the replication queue
//
function fifo_opendata()
    use replication!queue order cmdstate
    scatter name p_dataobject blank
return
